/*
#############################################################################

  Ch11p1_SimpleFire.cpp: a program that demonstrates the fire algorithm,
  without any annoying bells and/or whistles.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "Ch11p1_FirePalette.h"
#include "D3DHelperFuncs.h"
#include "Ch11p1_resource.h"

// A structure for our custom vertex type. 
struct CUSTOMVERTEX
{
  D3DXVECTOR3 position; // The position
  D3DCOLOR    color;    // The color
  FLOAT       tu, tv;   // The texture coordinates
};

const int TEXTURESIZE = 128; // size of the fire texture

// Our custom FVF, which describes our custom vertex structure
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1)

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  LPDIRECT3DVERTEXBUFFER8 m_pVB;
  DWORD        m_dwNumVertices;
  
  // Texture
  LPDIRECT3DTEXTURE8 m_pTexture;

  // Texture Palette
  PALETTEENTRY m_Palette[256];
  char m_strTextureSurfFormat[256];
  unsigned char m_cFireField[TEXTURESIZE*TEXTURESIZE];  // first fire array
  unsigned char m_cFireField2[TEXTURESIZE*TEXTURESIZE]; // second fire array
  unsigned char *m_pFireActive;   // we use these two pointers to flip
  unsigned char *m_pFireScratch;  // the active fire array back and forth.

protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

public:
  CMyD3DApplication();
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Ch11p1_SimpleFire");
  m_bUseDepthBuffer   = TRUE;

  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
  m_pVB              = NULL;
  m_dwNumVertices    = 6;
  m_pTexture         = NULL;
}

/****************************************************************************

 ProcessFire: this function processes our fire.  It takes two input buffers,
 the fire dimensions, and the cooling amount.  It calculates the new fire
 values from firefield1 and puts them into firefield2.

 ****************************************************************************/
void ProcessFire(unsigned char *firefield, unsigned char *firefield2, int firesizex, int firesizey, int coolamount)
{
  // loop through all the fire values...
  for (int y=0; y < firesizey; y++) {
    for (int x=0; x < firesizex-0; x++) {
      // these will store, temporarily, the fire values immediately to the
      // left, right, top, and bottom of this fire value.
      unsigned char firevalue_left, firevalue_right, firevalue_bottom, firevalue_top;
      int finalfirevalue;

      // x+-1 and y+-1 aren't always that simple; we must account for wrapping
      // around the horizontal edge (not vertical, however).
      // so we calculate x/y +- 1 and store them temporarily.
      int xplus1, xminus1, yplus1, yminus1;
      xplus1 = x+1; if (xplus1 >= firesizex) xplus1=0;
      xminus1= x-1; if (xminus1 < 0) xminus1 = firesizex-1;
      yplus1 = y+1; if (yplus1 >= firesizey) yplus1=firesizey-1;
      yminus1= y-1; if (yminus1 < 0) yminus1 = 0;
      
      // now we can get the fire values of the neighboring pixels
      firevalue_right = firefield[(y*firesizex)+xplus1];
      firevalue_left  = firefield[(y*firesizex)+xminus1];
      firevalue_bottom= firefield[((yplus1)*firesizex)+x];
      firevalue_top   = firefield[((yminus1)*firesizex)+x];

      // now, the most important part-  calculate the new fire value..
      finalfirevalue = (firevalue_left+firevalue_right+firevalue_top+firevalue_bottom)/4;

      // subtract a certain amount to simulate the fire "cooling."
      // this is where you'd apply your cooling map.
      finalfirevalue -= coolamount;

      // make sure that the subtraction of the coolamount didn't take us
      // below zero.
      if (finalfirevalue < 0) finalfirevalue = 0;

      // store the fire value on the scratch array, up one line from where
      // it originally was.  This simulates the flames rising.
      firefield2[((yminus1)*firesizex)+x] = finalfirevalue;
    }
  }

  // add fuel to the fire.  This particular fueling method is one of my
  // favorite; it creates little jet streams of flame.

  // we work in blocks of 2x1...
  for (int x=0; x < firesizex; x+=2) {
    // we add fuel only to the last row.
    int y=firesizey-1;
    
    // determine whether this particular spot gets fuel added or taken
    // away from it, by adding a number between (-31..31)
    int fuel = firefield[(y*firesizex)+x] + (rand() % 64) - 32;

    // we must be between 0-255.
    if (fuel > 255) fuel = 255;
    if (fuel < 0) fuel = 0;

    // apply the new fuel value to two adjacent pixels.  This helps reduce
    // the "dithering" effect that the fire is prone to.
    firefield2[(y*firesizex)+x] = (unsigned char)fuel;
    firefield2[(y*firesizex)+x+1] = (unsigned char)fuel;
  }
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  // set up the texture palette - in this program, we just grab it directly
  // from our palette header file.
  for (int q=0; q < 255; q++) {
    m_Palette[q].peRed   = g_FireRed[q];
    m_Palette[q].peGreen = g_FireGreen[q];
    m_Palette[q].peBlue  = g_FireBlue[q];
    m_Palette[q].peFlags = 0;
  }
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
  FLOAT fSecsPerFrame = m_fElapsedTime;

  // process the fire
  ProcessFire(m_pFireActive, m_pFireScratch, TEXTURESIZE, TEXTURESIZE, 1);

  // flip-flop the buffers.
  unsigned char *temp = m_pFireActive;
  m_pFireActive = m_pFireScratch;
  m_pFireScratch = temp;

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  HRESULT hr;
  
  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       0x000000, 1.0f, 0L );
  
  // lock texture
  D3DLOCKED_RECT lockedrect;
  ::ZeroMemory(&lockedrect, sizeof(lockedrect));
  
  if (FAILED(hr = m_pTexture->LockRect(0, &lockedrect, NULL, 0))) return(hr);
  
  // our texture surface is now locked, and we can use the pitch to traverse it.
  unsigned char *pSurfBits = static_cast<unsigned char *>(lockedrect.pBits);
  
  int index=0;
  for (int y=0; y < TEXTURESIZE; y++) {
    for (int x=0; x < TEXTURESIZE; x++) {
      // the fire value at this position determines the color of this texel
      pSurfBits[index++] = m_Palette[m_pFireActive[(y*TEXTURESIZE)+x]].peBlue; // blue
      pSurfBits[index++] = m_Palette[m_pFireActive[(y*TEXTURESIZE)+x]].peGreen; // green
      pSurfBits[index++] = m_Palette[m_pFireActive[(y*TEXTURESIZE)+x]].peRed; // red
      pSurfBits[index++] = m_Palette[m_pFireActive[(y*TEXTURESIZE)+x]].peFlags; // alpha
    }
    // next line
    index += lockedrect.Pitch - (TEXTURESIZE*4);
  }

  // unlock texture surface
  if (FAILED(hr = m_pTexture->UnlockRect(0))) return(hr);

  // Now that our fire texture's updated, we can begin rendering the scene
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    // draw our quad
    m_pd3dDevice->SetStreamSource( 0, m_pVB, sizeof(CUSTOMVERTEX) );
    m_pd3dDevice->SetVertexShader( D3DFVF_CUSTOMVERTEX );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 2 );

    // Output statistics
    m_pFont->DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
    m_pFont->DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats );
    m_pFont->DrawText( 2, 40, D3DCOLOR_ARGB(255,255,255,0), m_strTextureSurfFormat );

    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  HRESULT hr;

  m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();

  // Setup render states
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

  // Create vertex buffer
  {
    CUSTOMVERTEX* pVertices;

    if( FAILED( hr = m_pd3dDevice->CreateVertexBuffer( 6*sizeof(CUSTOMVERTEX),
                                                       0, D3DFVF_CUSTOMVERTEX,
                                                       D3DPOOL_MANAGED, &m_pVB ) ) )
      return hr;

    if( FAILED( hr = m_pVB->Lock( 0, m_dwNumVertices*sizeof(CUSTOMVERTEX), (BYTE**)&pVertices, 0 ) ) )
      return hr;

    // first triangle
    pVertices[0].position = D3DXVECTOR3(-1.0f, 1.0f, 0.0f);
    pVertices[0].color    = 0xffffffff;
    pVertices[0].tu       = 0.0f;
    pVertices[0].tv       = 0.0f;

    pVertices[1].position = D3DXVECTOR3(1.0f, 1.0f, 0.0f);
    pVertices[1].color    = 0xffffffff;
    pVertices[1].tu       = 1.0f;
    pVertices[1].tv       = 0.0f;

    pVertices[2].position = D3DXVECTOR3(1.0f, -1.0f, 0.0f);
    pVertices[2].color    = 0xffffffff;
    pVertices[2].tu       = 1.0f;
    pVertices[2].tv       = 1.0f;

    // second triangle
    pVertices[3].position = D3DXVECTOR3(-1.0f, 1.0f, 0.0f);
    pVertices[3].color    = 0xffffffff;
    pVertices[3].tu       = 0.0f;
    pVertices[3].tv       = 0.0f;

    pVertices[4].position = D3DXVECTOR3(1.0f, -1.0f, 0.0f);
    pVertices[4].color    = 0xffffffff;
    pVertices[4].tu       = 1.0f;
    pVertices[4].tv       = 1.0f;

    pVertices[5].position = D3DXVECTOR3(-1.0f, -1.0f, 0.0f);
    pVertices[5].color    = 0xffffffff;
    pVertices[5].tu       = 0.0f;
    pVertices[5].tv       = 1.0f;

    if( FAILED( hr = m_pVB->Unlock() ) )
      return hr;
  }

  // clear fire memory
  memset(m_cFireField, 0, TEXTURESIZE*TEXTURESIZE);
  ::ZeroMemory(m_cFireField2, TEXTURESIZE*TEXTURESIZE);
  
  // setup fire buffers
  m_pFireActive = m_cFireField;
  m_pFireScratch = m_cFireField2;

  // create fire texture
  {
    if (FAILED(hr = D3DXCreateTexture(m_pd3dDevice, TEXTURESIZE, TEXTURESIZE, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexture))) return(hr);
    //if (FAILED(hr = D3DXCreateTextureFromFile(m_pd3dDevice, "Ch8p1_Brick.bmp", &m_pTexture))) return(hr);

    LPDIRECT3DSURFACE8 pSurf;

    if (FAILED(hr = m_pTexture->GetSurfaceLevel(0, &pSurf))) return(hr);
    
    // get the texture's surface format, and store that in our variable for later display.
    D3DSURFACE_DESC surfdesc;
    pSurf->GetDesc(&surfdesc);

    strcpy(m_strTextureSurfFormat, SurfFormatToString(surfdesc.Format));

    SAFE_RELEASE(pSurf);
  }

  // Set up an orthagonal projection matrix, so we can render the entire
  // texture.
  D3DXMATRIX mat;
  D3DXMatrixOrthoLH(&mat, (float)TEXTURESIZE, (float)TEXTURESIZE, 
    0.0, 100.0);
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &mat );

  // this world matrix, combined with orthogonal projection, causes the 
  // texture to completely and exactly fill the rendering surface.
  D3DXMATRIX matWorld,matTrans,matScale;
  D3DXMatrixScaling(&matScale, (float)TEXTURESIZE/2.0f, (float)TEXTURESIZE/2.0f, 1.0);

  // move the quad left and up 0.5 units, so that the texels are perfectly
  // centered on the screen pixels.
  D3DXMatrixMultiply(&matWorld, &matScale, D3DXMatrixTranslation(&matTrans, -0.5f, -0.5f, 0));

  // our matrix is now finished.  Tell D3D to use it!
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );


  // set our fire texture active...
  m_pd3dDevice->SetTexture( 0, m_pTexture );

  // set up our texture stages for a simple texture copy...
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );

  // turning on linear filtering for our fire texture really helps out with
  // the image quality.
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  SAFE_RELEASE( m_pVB );
  SAFE_RELEASE( m_pTexture );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}